/*
** Module   : TableHints.java
** Abstract : manages hints read from a UAST hints file about a schema table
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20050713 ADD   @21702 Created initial version. Contains info to
**                               override name conversions at the table and
**                               field level, and to define new associations
**                               between tables, or to provide supplemental
**                               data for existing associations.
** 002 NVS 20050816 ADD   @22154 Added DROP attribute and DISPOSITION element
**                               definition and processing. These define how
**                               the table or table fields are imported.
**                               Added methods to query DROP attribute for
**                               table and fields, as well as DISPOSITIONs.
** 003 NVS 20050817 CHG   @22159 Changed methods returning dispositions from
**                               Set to List as these are readily accepted by
**                               putAnnotation() method.
** 004 ECF 20060828 CHG   @29081 Added 'encoded' attribute to field node.
**                               Indicates whether a character field must be
**                               stored base-64 encoded (and decoded on read).
** 005 ECF 20070303 CHG   @32266 Added 'temp' attribute to table node. This
**                               specifies an explicit name to apply to a
**                               temp/work-table defined like the permanent
**                               table for which the hint applies.
** 006 SVL 20080326 CHG   @37684 Added support for additional ("synthetic")
**                               indexes loaded from the hints file.  
** 007 VMN 20131114              Added support for custom conversion of fields with extents.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import org.w3c.dom.*;

import com.goldencode.util.StringHelper;

/**
 * Describes a variety of hint information for a single schema table.
 */
public class TableHints
{
   /** Field element of table and association elements */
   private static final String ELEM_FIELD = "field";
   
   /** Phrase element to define a customized name conversion */
   private static final String ELEM_PHRASE = "phrase";
   
   /** Disposition element to define a type of import processing */
   private static final String ELEM_DISP = "disposition";
   
   /** Relational association between this table and another */
   private static final String ELEM_ASSOCIATION = "association";
   
   /** Synthetic index element */
   private static final String ELEM_INDEX = "index";
   
   /** Field element of index */
   private static final String ELEM_INDEX_FIELD = "index-field";

   /** Element to define custom processing table fields with extent */
   private static final String ELEM_CUSTOM_EXTENT = "custom-extent";

   /**
    *  Element to define configurable, depending on customer preferences naming conversion
    *  fields with extent
    */
   private static final String ELEM_EXTENT_NAMES = "extent-names";

   /** Name attribute of table and field elements */
   private static final String ATTR_NAME = "name";
   
   /** Drop attribute of table and field elements */
   private static final String ATTR_DROP = "drop";
   
   /** Temp-table raw (pre-conversion) name attribute */
   private static final String ATTR_TEMP = "temp";
   
   /** Encoded attribute of field element */
   private static final String ATTR_ENCODED = "encoded";
   
   /** Unique attribute of index element */
   private static final String ATTR_UNIQUE = "unique";
   
   /** Direction attribute of index field element */
   private static final String ATTR_DIRECTION = "direction";
   
   /** Type attribute of association element */
   private static final String ATTR_TYPE = "type";
   
   /** Match attribute of phrase element */
   private static final String ATTR_MATCH = "match";
   
   /** Replace attribute of phrase element */
   private static final String ATTR_REPLACE = "replace";
   
   /** No-inherit attribute of phrase element */
   private static final String ATTR_NO_INHERIT = "no-inherit";
   
   /** References attribute of phrase element */
   private static final String ATTR_REFERENCES = "references";
   
   /** Foreign attribute of association elements */
   private static final String ATTR_FOREIGN = "foreign";
   
   /** Key under which join fields are stored in an association map */
   private static final String KEY_FIELDS = "fields";
   
   /** Key under which index fields are stored in the indexes map */
   private static final String KEY_INDEX_FIELDS = "index-fields";
   
   /** Value of type for many-to-one association */
   private static final String TYPE_MANY_TO_ONE = "many-to-one";
   
   /** Value of type for one-to-many association */
   private static final String TYPE_ONE_TO_MANY = "one-to-many";
   
   /** Value of type for one-to-one association */
   private static final String TYPE_ONE_TO_ONE = "one-to-one";
   
   /** Name conversion replacements for the table inherited by fields */
   private final Map inheritPhrases = new HashMap();
   
   /** Name conversion replacements for the table not inherited by fields */
   private final Map noInheritPhrases = new HashMap();
   
   /** Name conversion replacements for fields within table */
   private final Map fieldPhrases = new HashMap();
   
   /** Fields which require base-64 encoding for storage */
   private final Set encodedFields = new HashSet();
   
   /** Association definitions for this table */
   private final Map associations = new HashMap();
   
   /** Table dispositions for this table */
   private final Set tableDisp = new HashSet();
   
   /** Field drop flags for this table */
   private final Set fieldDrop = new HashSet();
   
   /** Field dispositions for this table */
   private final Map fieldDisp = new HashMap();
   
   /** Map containing information about synthetic indexes and their fields */
   private final Map indexes = new HashMap();

   /** Map containing information about custom conversion of fields with extents*/
   private Map<String, List <String>> customExtents = null;

   /** Name of table described by this hint instance */
   private final String name;
   
   /** Pre-conversion name of temp-table based upon underlying table */
   private final String tempTableName;
   
   /** Table drop flag for this table */
   private final boolean tableDrop;
   
   /**
    * Constructor which configures hint from an XML element.
    *
    * @param   table
    *          XML element describing the table hint. Corresponds with
    *          <code>table</code> tag within the hints file.
    */
   public TableHints(Element table)
   {
      NodeList list;
      int len;
      
      // Read name.
      name = table.getAttribute(ATTR_NAME);
      
      // Read table drop.
      String drop = table.getAttribute(ATTR_DROP);
      tableDrop = "true".equalsIgnoreCase(drop);
      
      // Read temp table raw name.
      String ttName = table.getAttribute(ATTR_TEMP);
      tempTableName = (ttName == null || ttName.trim().length() == 0
                       ? null
                       : ttName);
      
      // Read table and field phrases.
      list = table.getElementsByTagName(ELEM_PHRASE);
      len = list.getLength();
      for (int i = 0; i < len; i++)
      {
         Element phrase = (Element) list.item(i);
         Map map;
         Element parent = (Element) phrase.getParentNode();
         
         // Is parent a field element...
         if (parent.getTagName().equals(ELEM_FIELD))
         {
            String fieldName = parent.getAttribute(ATTR_NAME);
            map = (Map) fieldPhrases.get(fieldName);
            if (map == null)
            {
               map = new HashMap();
               fieldPhrases.put(fieldName, map);
            }
         }
         // ...or a table element?
         else
         {
            map = ("true".equals(phrase.getAttribute(ATTR_NO_INHERIT))
                   ? noInheritPhrases
                   : inheritPhrases);
         }
         
         map.put(phrase.getAttribute(ATTR_MATCH),
                 phrase.getAttribute(ATTR_REPLACE));
      }
      
      // Retrieve associations for table.
      list = table.getElementsByTagName(ELEM_ASSOCIATION);
      len = list.getLength();
      for (int i = 0; i < len; i++)
      {
         int size;
         Element assoc = (Element) list.item(i);
         
         // Store attributes in a map, each of which defines the properties
         // of an association and a simple key/value string pair.
         NamedNodeMap attrs = assoc.getAttributes();
         Map map = new HashMap();
         size = attrs.getLength();
         for (int j = 0; j < size; j++)
         {
            Attr attr = (Attr) attrs.item(j);
            map.put(attr.getName(), attr.getValue());
         }
         
         // Handle fields which define the join specially. These are
         // collected into a set and stored under the key "fields" in the
         // associations map.
         NodeList fields = assoc.getElementsByTagName(ELEM_FIELD);
         size = fields.getLength();
         Set set = null;
         for (int j = 0; j < size; j++)
         {
            Element field = (Element) fields.item(i);
            if (set == null)
            {
               set = new HashSet();
            }
            set.add(field.getAttribute(ATTR_NAME));
         }
         if (set != null)
         {
            map.put(KEY_FIELDS, set);
         }
         
         // Map associations by the name of the table they reference.
         associations.put(map.get(ATTR_REFERENCES), map);
      }
      
      // Read field drops and encoded fields.
      list = table.getElementsByTagName(ELEM_FIELD);
      len = list.getLength();
      for (int i = 0; i < len; i++)
      {
         Element field = (Element) list.item(i);
         String fieldName = field.getAttribute(ATTR_NAME);
         
         drop = field.getAttribute(ATTR_DROP);
         if ("true".equalsIgnoreCase(drop))
         {
            fieldDrop.add(fieldName);
         }
         
         String encoded = field.getAttribute(ATTR_ENCODED);
         if ("true".equalsIgnoreCase(encoded))
         {
            encodedFields.add(fieldName);
         }
      }
      
      // Read table and field dispositions.
      list = table.getElementsByTagName(ELEM_DISP);
      len = list.getLength();
      for (int i = 0; i < len; i++)
      {
         Element disp = (Element) list.item(i);
         Set set;
         Element parent = (Element) disp.getParentNode();
         
         // Is parent a field element...
         if (parent.getTagName().equals(ELEM_FIELD))
         {
            String fieldName = parent.getAttribute(ATTR_NAME);
            set = (Set) fieldDisp.get(fieldName);
            if (set == null)
            {
               set = new HashSet();
               fieldDisp.put(fieldName, set);
            }
         }
         // ...or a table element?
         else
         {
            set = tableDisp;
         }
         
         set.add(disp.getAttribute(ATTR_NAME));
      }

      /*
       * Read custom extent elements into the map like field-name -> list of extent property
       * names.
       *
       * Using this hint for fields with extent:
       *
       * a) customExtents == null
       * It means that this table has no extent fields with custom conversion.
       *
       * b) customExtents.get(field-name) == null
       * It means that this field has no custom conversion.
       *
       * c) customExtents.get(field-name) returns empty list
       * It means that this field has custom conversion with default naming conversion
       * (e.g., field(1)-->field_1, field(2)-->field2, ... field(N)-->fieldN).
       *
       * d) customExtents.get(field-name) returns non-empty list
       * It means that this field has custom conversion with configurable, depending on customer
       * preferences (e.g., allowing individual indexes to be named explicitly, as in,
       * hours(1)-->monday_hours, hours(2)-->tuesday_hours, etc.).
       * TODO: check in conversion time sizes of this list and extent value
       */
      list = table.getElementsByTagName(ELEM_CUSTOM_EXTENT);
      len = list.getLength();
      if (len > 0)
      {
         customExtents = new HashMap<String, List <String>>();
         for (int i = 0; i < len; i++)
         {
            Element extent = (Element) list.item(i);
            NodeList fieldsList = extent.getElementsByTagName(ELEM_FIELD);
            int fieldsListLen = fieldsList.getLength();
            if (fieldsListLen == 0)
            {
               customExtents.clear();
               break;
            }
            else
            {
               for (int k = 0; k < fieldsListLen; k++)
               {
                  Element field = (Element) fieldsList.item(k);
                  NodeList extentNames = field.getElementsByTagName(ELEM_EXTENT_NAMES);
                  List<String> fieldNames;
                  if (extentNames.getLength() == 0)
                  {
                     fieldNames = new ArrayList<String>();
                  }
                  else
                  {
                     Element extentNamesElement = (Element) extentNames.item(0);
                     String names = extentNamesElement.getTextContent();
                     fieldNames = StringHelper.parseTokens(names, ",", true, true, true);
                  }
                  String fieldName = field.getAttribute("name");
                  customExtents.put(fieldName, fieldNames);
               }
            }
         }
      }

      /*
       * Read table synthetic indexes into the map like
       * index-name -> 'unique' -> true/false
       *               'index-fields' -> field-name -> 'direction' -> asc/desc
       *                                 field-name2 -> ...
       * index-name2 -> ...
       */ 
      list = table.getElementsByTagName(ELEM_INDEX);
      len = list.getLength();
      for (int i = 0; i < len; i++)
      {
         Element index = (Element) list.item(i);
         String indexName = index.getAttribute(ATTR_NAME);
         
         Map map = new HashMap();
         map.put(ATTR_UNIQUE, index.getAttribute(ATTR_UNIQUE));
         
         NodeList fields = index.getElementsByTagName(ELEM_INDEX_FIELD);
         Map fieldsMap = new HashMap();
         int fieldsLen = fields.getLength();
         for (int j = 0; j < fieldsLen; j++)
         {
            Element field = (Element) fields.item(j);
            String fieldName = field.getAttribute(ATTR_NAME);
            
            Map propertiesMap = new HashMap();
            propertiesMap.put(ATTR_DIRECTION, field.getAttribute(ATTR_DIRECTION));
            
            fieldsMap.put(fieldName,propertiesMap);
         }
         
         map.put(KEY_INDEX_FIELDS, fieldsMap);
         
         indexes.put(indexName,map);
      }
   }
   
   /**
    * Get the name of the table described by this hint.
    *
    * @return  Table name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Get the pre-conversion name of a temp table which is based upon the
    * underlying table.
    *
    * @return  Raw temp table name to be passed to the name converter, or
    *          <code>null</code> if no explicit temp-table name exists.
    */
   public String getTempTableName()
   {
      return tempTableName;
   }
   
   /**
    * Get the custom replacement map (if any) used for name conversion for
    * this table. This is a combination of all inheritable and
    * non-inheritable phrases for this table.
    *
    * @return  Replacement map for this table, which may be empty if no
    *          phrase hints are defined, but will not be <code>null</code>.
    */
   public Map getNameReplacements()
   {
      Map map = new HashMap();
      map.putAll(inheritPhrases);
      map.putAll(noInheritPhrases);
      
      return map;
   }
   
   /**
    * Get the custom replacement map (if any) used for name conversion for
    * the specified field in this table. This is a combination of all
    * inheritable phrases for this table and the phrases defined specifically
    * for the named field.
    *
    * @param   field
    *          Name of the field for which to retrieve replacement phrases.
    *
    * @return  Replacement map for this table, which may be empty if no
    *          phrase hints are defined, but will not be <code>null</code>.
    */
   public Map getNameReplacements(String field)
   {
      Map map = new HashMap();
      map.putAll(inheritPhrases);
      Map fieldMap = (Map) fieldPhrases.get(field);
      if (fieldMap != null)
      {
         map.putAll(fieldMap);
      }
      
      return map;
   }
   
   /**
    * Report whether the specified field must be base-64 encoded for data
    * storage (and decoded upon retrieval).
    *
    * @param   field
    *          Name of the field to test for the &quot;encoded&quot;
    *          attribute.
    *
    * @return  <code>true</code> if data must be base-64 encoded/decoded.
    */
   public boolean isFieldEncoded(String field)
   {
      return encodedFields.contains(field);
   }
   
   /**
    * Get an iterator on the names of all tables referenced by relational
    * associations from this table, in no particular order.
    *
    * @return  Iterator on referenced table names.
    */
   public Iterator associatedTables()
   {
      return associations.keySet().iterator();
   }
   
   /**
    * Get the schema parser token type associated with the association (if
    * any) described by this hint between this table and the table it
    * references. Valid token types are:
    * <ul>
    * <li><code>MANY_TO_ONE</code>
    * <li><code>ONE_TO_MANY</code>
    * <li><code>ONE_TO_ONE</code>
    * </ul>
    *
    * @param   references
    *          Name of table which the association references.
    *
    * @return  One of the above token types, or <code>-1</code> if either the
    *          association does not exist, or an unrecognized association
    *          type was found.
    */
   public int getAssociationType(String references)
   {
      int ttype = -1;
      Map map = (Map) associations.get(references);
      if (map != null)
      {
         String type = (String) map.get(ATTR_TYPE);
         if (TYPE_MANY_TO_ONE.equals(type))
         {
            ttype = SchemaParserTokenTypes.MANY_TO_ONE;
         }
         else if (TYPE_ONE_TO_MANY.equals(type))
         {
            ttype = SchemaParserTokenTypes.ONE_TO_MANY;
         }
         else if (TYPE_ONE_TO_ONE.equals(type))
         {
            ttype = SchemaParserTokenTypes.ONE_TO_ONE;
         }
      }
      
      return ttype;
   }
   
   /**
    * Indicate whether the table described by this hint represents the
    * foreign key end of an association with the specified table. This is
    * only relevant for one-to-one associations.
    *
    * @param   references
    *          Name of table which the association references.
    *
    * @return  <code>true</code> if this table contains a foreign key for
    *          the referenced table. <code>false</code> if not, or if no
    *          association was found with table <code>references</code>.
    */
   public boolean isForeignAssociation(String references)
   {
      Map map = (Map) associations.get(references);
      
      return (map != null && "true".equals(map.get(ATTR_FOREIGN)));
   }

   /**
    * Get the list of extent property names for given field with extent:
    * several property names in case of custom extent conversion.
    *
    * @param   fieldName
    *          Field name.
    *
    * @return  list of property names.
    */
   public List<String> getCustomFieldExtent(String fieldName)
   {
      if (customExtents == null)
      {
         return null;
      }
      else if (customExtents.size() == 0)
      {
         return new ArrayList<String>();
      }
      else
      {
         return customExtents.get(fieldName);
      }
   }
   
   /**
    * Get the set of field names which are used to form a relational
    * association between the table described by this hint and the table
    * it references.
    *
    * @param   references
    *          Name of table which the association references.
    *
    * @return  Set of fields, or <code>null</code> if this hint does not
    *          describe the referenced association or describes only the
    *          basic attributes of the association and not the joining
    *          fields.
    */
   public Set getAssociatedFields(String references)
   {
      Map map = (Map) associations.get(references);
      
      return (Set) (map == null ? null : map.get(KEY_FIELDS));
   }
   
   /**
    * Indicate whether the table described by this hint is to be dropped 
    * during the database import.
    *
    * @return  <code>true</code> if this table is to be dropped.
    */
   public boolean isTableDropped()
   {
      return tableDrop;
   }
   
   /**
    * Indicate whether the field of the table described by this hint is  
    * to be dropped during the database import.
    *
    * @param field
    *        field name to query.
    *
    * @return  <code>true</code> if this table is to be dropped.
    */
   public boolean isFieldDropped(String field)
   {
      return fieldDrop.contains(field);
   }
   
   /**
    * Get the list of table dispositions defined for the table described 
    * by this hint. 
    *
    * @return  List of dispositions, possibly empty.
    */
   public List getTableDispositions()
   {
      List list = new LinkedList();
      Iterator iter = tableDisp.iterator();
      Object disp = null;
      
      while(iter.hasNext())
      {
         disp = iter.next();
         list.add(disp);
      }
      
      return list;
   }
   
   /**
    * Get the list of field dispositions defined for the table described 
    * by this hint. 
    *
    * @param field
    *        field name to get the dispositions for.
    *
    * @return  List of dispositions, possibly empty.
    */
   public List getFieldDispositions(String field)
   {
      List list = new LinkedList();
      Set set = (Set) fieldDisp.get(field);
      
      if (set == null)
         return list;
      
      Iterator iter = set.iterator();
      Object disp = null;
      
      while(iter.hasNext())
      {
         disp = iter.next();
         list.add(disp);
      }
      
      return list;
   }
   
   /**
    * Get the set of synthetic index names.
    *
    * @return Set of index names, possibly empty.
    */
   public Set getIndexNames()
   {
      return indexes.keySet();
   }
   
   /**
    * Determine whether the specified field in the specified index is sorted
    * in the descending order.
    *
    * @param  indexName
    *         The name of the index that contains the required field.
    *
    * @param  fieldName
    *         Field name to query.
    *
    * @return <code>true</code> if the required field is sorted in the
    *         descending order. <code>false</code> otherwise.
    */
   public boolean isIndexFieldDescending(String indexName, String fieldName)
   {
      Map indexMap = (Map) indexes.get(indexName);
      Map fieldMap =
            (Map)((Map)indexMap.get(KEY_INDEX_FIELDS)).get(fieldName);
      return "desc".equalsIgnoreCase((String)fieldMap.get(ATTR_DIRECTION)) ||
             "descend".equalsIgnoreCase((String)fieldMap.get(ATTR_DIRECTION));
   }
   
   /**
    * Determine whether the specified index is unique. 
    *
    * @param  indexName
    *         Index to query.
    *
    * @return <code>true</code> if the required index is unique.
    *         <code>false</code> otherwise.
    */
   public boolean isIndexUnique(String indexName)
   {
      Map indexMap = (Map) indexes.get(indexName);
      return "true".equalsIgnoreCase((String)indexMap.get(ATTR_UNIQUE));
   }
   
   /**
    * Get the set of field names of the specified index.
    *
    * @param  indexName
    *         The name of the index to get the field names for.
    *
    * @return Set of index field names, possibly empty.
    */
   public Set getIndexFieldNames(String indexName)
   {
      Map indexMap = (Map) indexes.get(indexName);
      return ((Map)indexMap.get(KEY_INDEX_FIELDS)).keySet();
   }
}
